//+------------------------------------------------------------------+
//|                                                    ZoGIF library |
//|                                            Copyright 2015, Zorro |
//|                              https://www.mql5.com/ru/users/zorro |
//+------------------------------------------------------------------+
#property copyright "Copyright 2015, Zorro"
#property link      "https://www.mql5.com/ru/users/zorro"
#property version   "1.00"
#property library

#define Zo_BUILD_LIB

#include <Files/FileBin.mqh>
#include <ZoLib/img/ZoGIF.mqh>
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
#define CopyArray2Struct_HELPER(s) struct __FUNCTION__##__LINE__##s { uchar data[sizeof(s)]; }
template<typename T>
bool CopyArray2Struct(T &s,const uchar &data[],int start)
  {
   CopyArray2Struct_HELPER(T) dst;
//---
   if(ArrayCopy(dst.data,data,0,start,sizeof(T))!=sizeof(T))
      return(false);
//---
   s=(T)dst;
   return(true);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
struct SGIFImage
  {
   ushort            image_left;
   ushort            image_top;
   ushort            image_width;
   ushort            image_height;
   uchar             image_info;
   int               delay;
   uint              data[];
  };  
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
struct SGIFHeader
  {
   uchar             signatute[6];
   ushort            image_width;
   ushort            image_height;
   uchar             image_info;
   uchar             image_background;
   uchar             image_ratio;
  };  
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
struct SGIFTable
  {
   uchar             code;
   ushort            prev;
   ushort            next;
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
struct SGIFControl
  {
   uchar             ctrl;
   ushort            delay;
   uchar             transparent;
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
struct SGIFStream
  {
   uint              size;
   uchar             data[],byte;
   uint              offset;
   //--- for image unpack
   uint              bit_size;
   uint              pbit_size;
   uint              x,y,iw,ih,pass,row,rem;
   uint              index,block,read,pal_size,pal_offset;
   SGIFControl       ctrl;
   SGIFTable         table[];
   uchar             dispose;
   bool              is_ctrl;
   //---
   ushort            NextCode();
   uchar             NextByte();
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CGifImage
  {
   SGIFHeader        m_gif_header;
   uchar             m_gif_palette[];
   SGIFImage         m_gif_images[];
   SGIFStream        m_gif_stream;

public:
   bool              Load(ZoGIF &gif,const uchar &data[],int start,int count);
   uint              FramesCount(void) const { return(ArraySize(m_gif_images)); }
   bool              FrameToResource(string res_name,uint frame);
   uint              FrameDelay(uint frame);   
   bool              GetGIF(ZoGIF &gif);

private:
   bool              LoadExtention();
   bool              LoadImage();
   void              SetPixel(uchar code);
  };
//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
bool ZoGIFLoad(/* OUT */ZoGIF &gif,/* IN */const string path) export
  {
   CFileBin file;  
   uchar    data[];
//--- open file
   if(!file.Open(path,FILE_READ))
      return(false);
//--- read file
   if(file.ReadArray(data)==0)
      return(false);
//--- parse GIF
   return(ZoGIFLoad(gif,data));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+  
bool ZoGIFLoad(/* OUT */ZoGIF &gif,/* IN */const uchar &data[],/* IN OPT */int start=0,/* IN OPT */int count=WHOLE_ARRAY) export
  {
//--- we should to calculate
   if(count==WHOLE_ARRAY)
      count=ArraySize(data)-start;
//--- checkup counter
   if(count<0)
      return(false);      
//--- load gif image
   CGifImage img;
   return(img.Load(gif,data,start,count));
  }
//+------------------------------------------------------------------+
//| Common Load GIF function                                         |
//+------------------------------------------------------------------+
bool CGifImage::Load(ZoGIF &gif,const uchar &data[],int start,int count)
  {
//--- checkup stream size
   if(count<sizeof(m_gif_header))
      return(false);      
//--- read GIF header from stream
   if(!CopyArray2Struct(m_gif_header,data,start))
      return(false);   
//--- header readed
   start+=(int)sizeof(m_gif_header);
   count-=(int)sizeof(m_gif_header);
//--- global color map used
   if((m_gif_header.image_info&0x80)==0x80)
     {
      uchar color_res =(m_gif_header.image_info&0x70)>>4;
      uchar pixel_size=(m_gif_header.image_info&0x07)+1;
      int   gcm_size  =(1<<pixel_size)*3;
      //--- checkup stream size
      if(count<gcm_size)
         return(false);     
      //--- read gloabal palette from stream
      if(ArrayCopy(m_gif_palette,data,0,start,gcm_size)!=gcm_size)
         return(false);
      //--- global palette readed
      start+=gcm_size;
      count-=gcm_size;
     }
//---
   m_gif_stream.offset  =0;
   m_gif_stream.index   =0;
   m_gif_stream.size    =count;
   m_gif_stream.dispose =2;
   m_gif_stream.is_ctrl =false;
   if(ArrayResize(m_gif_stream.table,4096)!=4096)
      return(false);
   if(ArrayCopy(m_gif_stream.data,data,0,start,count)!=count)
      return(false);
//---
   for(;m_gif_stream.offset<m_gif_stream.size;m_gif_stream.offset++)
     {
      switch(m_gif_stream.data[m_gif_stream.offset])
        {
         case 0x00:
         case 0x3B:     // trailer
            break;
         case 0x2C:     // an image
            if(!LoadImage())
               return(false);
            break;
         case 0x21:     // a data
            if(!LoadExtention())
               return(false);
            break;
         default:
            return(false);
        }
     }
//---
   return(GetGIF(gif));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool CGifImage::LoadExtention()
  {
   if(++m_gif_stream.offset>=m_gif_stream.size)
      return(false);
//--- parse extention
   switch(m_gif_stream.data[m_gif_stream.offset])
     {
      case 0x01:     // plain text
         m_gif_stream.offset+=13;
         do
           {
            if(++m_gif_stream.offset>=m_gif_stream.size)
               return(false);
           }
         while(m_gif_stream.data[m_gif_stream.offset]!=0);
         break;
      case 0xF9:     // graphic control
         m_gif_stream.is_ctrl          =true;
         m_gif_stream.ctrl.ctrl        =m_gif_stream.data[m_gif_stream.offset+2];
         m_gif_stream.ctrl.delay       =(m_gif_stream.data[m_gif_stream.offset+4]<<8)+m_gif_stream.data[m_gif_stream.offset+3];
         m_gif_stream.ctrl.transparent =m_gif_stream.data[m_gif_stream.offset+5];
         m_gif_stream.offset          +=6;
         return(m_gif_stream.offset<m_gif_stream.size);
      case 0xFE:     // comment
         do
           {
            if(++m_gif_stream.offset>=m_gif_stream.size)
               return(false);
           }
         while(m_gif_stream.data[m_gif_stream.offset]!=0);
         break;
      case 0xFF:     // application
         m_gif_stream.offset+=12;
         do
           {
            if(++m_gif_stream.offset>=m_gif_stream.size)
               return(false);
           }
         while(m_gif_stream.data[m_gif_stream.offset]!=0);
         //---    application data
         for(;m_gif_stream.offset<m_gif_stream.size;m_gif_stream.offset++)
           {
            if(m_gif_stream.data[m_gif_stream.offset+1]!=0)
               return(true);
           }
         //---                  
         return(false);
     }
//---
   return(true);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool CGifImage::LoadImage()
  {
   if(m_gif_stream.offset+sizeof(SGIFImage)>=m_gif_stream.size)
      return(false);
//---
   if(ArrayResize(m_gif_images,m_gif_stream.index+1,32)!=m_gif_stream.index+1)
      return(false);
//---
   m_gif_stream.pal_offset=0;
   m_gif_stream.pal_size  =ArraySize(m_gif_palette);
//--- skip the separator ','
   m_gif_stream.offset++;
//--- load image info
   m_gif_images[m_gif_stream.index].image_left  =m_gif_stream.data[m_gif_stream.offset++]+(m_gif_stream.data[m_gif_stream.offset++]<<8);
   m_gif_images[m_gif_stream.index].image_top   =m_gif_stream.data[m_gif_stream.offset++]+(m_gif_stream.data[m_gif_stream.offset++]<<8);
   m_gif_images[m_gif_stream.index].image_width =m_gif_stream.data[m_gif_stream.offset++]+(m_gif_stream.data[m_gif_stream.offset++]<<8);
   m_gif_images[m_gif_stream.index].image_height=m_gif_stream.data[m_gif_stream.offset++]+(m_gif_stream.data[m_gif_stream.offset++]<<8);
   m_gif_images[m_gif_stream.index].image_info  =m_gif_stream.data[m_gif_stream.offset++];
   m_gif_images[m_gif_stream.index].delay       =100;
   if(m_gif_stream.is_ctrl)
      m_gif_images[m_gif_stream.index].delay=m_gif_stream.ctrl.delay;
//--- image with local palette
   if((m_gif_images[m_gif_stream.index].image_info&0x80)!=0)
     {
      m_gif_stream.pal_size  =1<<((m_gif_images[m_gif_stream.index].image_info&7)+1);
      m_gif_stream.pal_offset=m_gif_stream.offset;
      m_gif_stream.offset   +=m_gif_stream.pal_size*3;
     }

   m_gif_stream.bit_size=m_gif_stream.data[m_gif_stream.offset++];
//---   
   uint tbl_size     =(1<<m_gif_stream.bit_size)+2;
   uint beg_size     =tbl_size;
   uint end_code     =tbl_size-1;     // EOI = 2**N+1
   uint clr_code     =tbl_size-2;     // CC  = 2**N
   ushort code,prev  =0;

   m_gif_stream.bit_size++;
   m_gif_stream.pbit_size=m_gif_stream.bit_size;
   m_gif_stream.rem  =0;
   m_gif_stream.byte =0;
   m_gif_stream.block=0;
   m_gif_stream.read =1;
   m_gif_stream.x    =0;
   m_gif_stream.y    =0;
   m_gif_stream.pass =1;
   m_gif_stream.row  =0;
   m_gif_stream.iw   =m_gif_images[m_gif_stream.index].image_width;
   m_gif_stream.ih   =m_gif_images[m_gif_stream.index].image_height;

   uint tmp;

   switch(m_gif_stream.dispose)
     {
      default:
      case 1:
      case 3:
            ArrayCopy(m_gif_images[m_gif_stream.index].data,m_gif_images[m_gif_stream.index-1].data);
            break;            
      case 2:
         tmp=m_gif_header.image_width*m_gif_header.image_height;
         if(ArrayResize(m_gif_images[m_gif_stream.index].data,tmp)!=tmp)
            return(false);
         //---
         if((m_gif_stream.ctrl.ctrl&1)==0)
           {
            tmp=0xFF000000;
            if(m_gif_stream.pal_offset!=0)
              {
               tmp|=m_gif_stream.data[m_gif_stream.pal_offset+m_gif_header.image_background*3+0]<<16;
               tmp|=m_gif_stream.data[m_gif_stream.pal_offset+m_gif_header.image_background*3+1]<< 8;
               tmp|=m_gif_stream.data[m_gif_stream.pal_offset+m_gif_header.image_background*3+2];
              }
            else
              {
               tmp|=m_gif_palette[m_gif_header.image_background*3+0]<<16;
               tmp|=m_gif_palette[m_gif_header.image_background*3+1]<< 8;
               tmp|=m_gif_palette[m_gif_header.image_background*3+2];
              }
           }
         else
            tmp=0x00000000;
         //---
         ArrayInitialize(m_gif_images[m_gif_stream.index].data,tmp);
         break;
     }

   ZeroMemory(m_gif_stream.table);
//--- unpack image
   while((code=m_gif_stream.NextCode())!=end_code)
     {
      if(code==clr_code)
        {
         m_gif_stream.bit_size=m_gif_stream.pbit_size;
         tbl_size             =beg_size;
         prev                 =m_gif_stream.NextCode();
         //---
         if(prev>tbl_size)
            return(false);
         //---
         SetPixel((uchar)prev);
         continue;
        }

      ushort c1,c2=0;
      //---
      if(code<tbl_size)
        {
         c1=code;
         //---
         while(c1>=beg_size)
           {
            m_gif_stream.table[c1].next=c2;
            c2=c1;
            c1=m_gif_stream.table[c1].prev;
            if(c1>=c2)
               return(false);
           }
         //---
         SetPixel((uchar)c1);
         //----
         while(c2!=0)
           {
            SetPixel(m_gif_stream.table[c2].code);
            c2=m_gif_stream.table[c2].next;
           }
        }
      else
        {
         c1=prev;
         //---
         while(c1>=beg_size)
           {
            m_gif_stream.table[c1].next=c2;
            c2=c1;
            c1=m_gif_stream.table[c1].prev;
            if(c1>=c2)
               return(false);
           }

         SetPixel((uchar)c1);

         while(c2!=0)
           {
            SetPixel(m_gif_stream.table[c2].code);
            c2=m_gif_stream.table[c2].next;
           }

         SetPixel((uchar)c1);
        }
      //---
      m_gif_stream.table[tbl_size].code=(uchar)c1;
      m_gif_stream.table[tbl_size].prev=prev;
      tbl_size++;
      //---
      if(tbl_size==((uint)1<<m_gif_stream.bit_size))
        {
         m_gif_stream.bit_size++;
         //---
         if(m_gif_stream.bit_size>12)
            m_gif_stream.bit_size=12;
        }
      //---
      prev=code;
     }
//---
   m_gif_stream.index++;
//---
   if(m_gif_stream.is_ctrl)
      m_gif_stream.dispose=uchar((m_gif_stream.ctrl.ctrl&0x1F)>>2);
//---
   m_gif_stream.is_ctrl=false;
//---
   return(true);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CGifImage::SetPixel(uchar code)
  {
   uint y;
   uint x;
//---
   if((m_gif_images[m_gif_stream.index].image_info&0x40)!=0)
     {
      if(m_gif_stream.x==m_gif_stream.iw)
        {
         m_gif_stream.x=0;
         if(m_gif_stream.pass==1) m_gif_stream.row+=8;
         if(m_gif_stream.pass==2) m_gif_stream.row+=8;
         if(m_gif_stream.pass==3) m_gif_stream.row+=4;
         if(m_gif_stream.pass==4) m_gif_stream.row+=2;

         if(m_gif_stream.row>=m_gif_stream.ih)
           {
            m_gif_stream.pass+=1;
            m_gif_stream.row=16>>m_gif_stream.pass;
           }
        }
      //---
      y=m_gif_stream.row;
     }
   else
     {
      if(m_gif_stream.x==m_gif_stream.iw)
        {
         m_gif_stream.x=0;
         m_gif_stream.y++;
        }
      //---
      y=m_gif_stream.y;
     }
//---
   y+=m_gif_images[m_gif_stream.index].image_top;
   x =m_gif_stream.x+m_gif_images[m_gif_stream.index].image_left;
   m_gif_stream.x++;
//---
   if(x<m_gif_header.image_width && y<m_gif_header.image_height)
     {
      if((m_gif_stream.ctrl.ctrl&1)==0 || code!=m_gif_stream.ctrl.transparent)
        {
         uint clr=0xFF000000;
         if(m_gif_stream.pal_offset!=0)
           {
            clr|=m_gif_stream.data[m_gif_stream.pal_offset+code*3+0]<<16;
            clr|=m_gif_stream.data[m_gif_stream.pal_offset+code*3+1]<< 8;
            clr|=m_gif_stream.data[m_gif_stream.pal_offset+code*3+2];
           }
         else
           {
            clr|=m_gif_palette[code*3+0]<<16;
            clr|=m_gif_palette[code*3+1]<< 8;
            clr|=m_gif_palette[code*3+2];
           }
         //---
         m_gif_images[m_gif_stream.index].data[y*m_gif_header.image_width+x]=clr;
        }
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
ushort SGIFStream::NextCode(void)
  {
   uint  tmp1;
   uchar tmp=1;
//---
   if(rem>=bit_size)
     {
      tmp=uchar(1<<bit_size);
      tmp--;
      tmp1=byte&tmp;
      byte=uchar(byte>>bit_size);
      rem-=bit_size;
     }
   else
     {
      tmp=uchar(1<<rem);
      tmp--;
      tmp1=byte;
      byte=NextByte();
      tmp =1;
      if(8>=(bit_size-rem))
        {
         tmp=uchar(1<<(bit_size-rem));
         tmp--;
         tmp1=(uint(byte&tmp)<<rem)+tmp1;
         byte=uchar(byte>>(bit_size-rem));
         rem =8-(bit_size-rem);
        }
      else
        {
         tmp1 =((uint(byte))<<rem)+tmp1;
         byte =NextByte();
         tmp  =uchar(1<<(bit_size-rem-8));
         tmp--;
         tmp1 =(uint(byte&tmp)<<(rem+8))+tmp1;
         byte =uchar(byte>>(bit_size-rem-8));
         rem  =8-(bit_size-rem-8);
        }
     }
//---
   return ushort(tmp1);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
uchar SGIFStream::NextByte(void)
  {
   if(read>=block)
     {
      block     =data[offset++];
      read      =0;

      if(block==0)
        {
         offset--;
         return 0xFF;
        }
     }
//---
   read++;
//---
   return data[offset++];
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool CGifImage::FrameToResource(string res_name,uint frame)
  {
   if(frame>=FramesCount())
      return(false);
//---
   return(ResourceCreate(res_name,m_gif_images[frame].data,m_gif_header.image_width,m_gif_header.image_height,0,0,
          m_gif_header.image_width,COLOR_FORMAT_ARGB_NORMALIZE));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
uint CGifImage::FrameDelay(uint frame)
  {
   if(frame>=FramesCount())
      return(false);
//---
   return(m_gif_images[frame].delay*10);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool CGifImage::GetGIF(ZoGIF &gif)
  {
   bool res=true;
//---
   gif.width      =m_gif_header.image_width;
   gif.height     =m_gif_header.image_height;   
   gif.frame_count=ArraySize(m_gif_images);
//--- does we have any frames?
   if(gif.frame_count>0)
     {
      if(ArrayResize(gif.delay,gif.frame_count)!=gif.frame_count)
         res=false;
      if(ArrayResize(gif.frame,gif.frame_count)!=gif.frame_count)
         res=false;
      //---
      if(res)
        {
         for(int i=0;i<gif.frame_count;i++)
           {
            gif.delay[i]=m_gif_images[i].delay*10;          
            if(ArrayCopy(gif.frame[i].image,m_gif_images[i].data)!=ArraySize(m_gif_images[i].data))
              {
               res=false;
               break;
              }
           }           
        }
     }
//---
   if(!res || gif.frame_count<=0)
     {
      ArrayFree(gif.delay);
      ArrayFree(gif.frame);
      //---
      if(!res)
        {
         gif.width      =0;
         gif.height     =0;
         gif.frame_count=0;
        }
     }
//---
   return(res);
  }
//+------------------------------------------------------------------+
